@@ -3,23 +3,22 @@
 
 """Post-processing image optimization of OCR PDFs."""
 
-
 from __future__ import annotations
 
 import logging
 import sys
 import tempfile
 import threading
-from collections import defaultdict
-from collections.abc import Iterator, MutableSet, Sequence
+from collections.abc import Callable, Iterator, MutableSet, Sequence
 from os import fspath
 from pathlib import Path
-from typing import Any, Callable, NamedTuple, NewType
-from warnings import warn
+from typing import Any, NamedTuple, NewType
 from zlib import compress
 
 import img2pdf
+from packaging.version import Version
 from pikepdf import (
+    Array,
     Dictionary,
     Name,
     Object,
@@ -30,10 +29,11 @@ from pikepdf import (
     Stream,
     UnsupportedImageTypeError,
 )
+from pikepdf.models.image import HifiPrintImageNotTranscodableError
 from PIL import Image
 
 from ocrmypdf._concurrent import Executor, SerialExecutor
-from ocrmypdf._exec import jbig2enc, pngquant
+from ocrmypdf._exec import ghostscript, jbig2enc, pngquant
 from ocrmypdf._jobcontext import PdfContext
 from ocrmypdf._progressbar import ProgressBar
 from ocrmypdf.exceptions import OutputFileAccessError
@@ -43,6 +43,7 @@ log = logging.getLogger(__name__)
 
 DEFAULT_JPEG_QUALITY = 75
 DEFAULT_PNG_QUALITY = 70
+FLATE_JPEG_THRESHOLD = 10000
 
 
 Xref = NewType('Xref', int)
@@ -76,10 +77,15 @@ def extract_image_filter(
     """Determine if an image is extractable."""
     if image.Subtype != Name.Image:
         return None
-    if image.Length < 100:
+    if not isinstance(image.Length, int) or image.Length < 100:
         log.debug(f"xref {xref}: skipping image with small stream size")
         return None
-    if image.Width < 8 or image.Height < 8:  # Issue 732
+    if (
+        not isinstance(image.Width, int)
+        or not isinstance(image.Height, int)
+        or image.Width < 8
+        or image.Height < 8
+    ):  # Issue 732
         log.debug(f"xref {xref}: skipping image with unusually small dimensions")
         return None
 
@@ -91,6 +97,7 @@ def extract_image_filter(
         if (
             len(pim.filter_decodeparms) == 2
             and first_filtdp[0] == Name.FlateDecode
+            and first_filtdp[1] is not None
             and first_filtdp[1].get(Name.Predictor, 1) == 1
             and second_filtdp[0] == Name.DCTDecode
             and not second_filtdp[1]
@@ -121,6 +128,13 @@ def extract_image_filter(
     if Name.Decode in image:
         log.debug(f"xref {xref}: skipping image with Decode table")
         return None  # Don't mess with custom Decode tables
+    if image.get(Name.SMask, Dictionary()).get(Name.Matte, None) is not None:
+        # https://github.com/ocrmypdf/OCRmyPDF/issues/1536
+        # Do not attempt to optimize images that have a SMask with a Matte.
+        # That means alpha channel pre-blending is used, and we're not prepared
+        # to deal with the complexities of that.
+        log.debug(f"xref {xref}: skipping image whose SMask has Matte")
+        return None
 
     return pim, filtdp
 
@@ -154,7 +168,17 @@ def extract_image_jbig2(
                 imgname = root / f'{xref:08d}'
                 with imgname.open('wb') as f:
                     ext = pim.extract_to(stream=f)
-                imgname.rename(imgname.with_suffix(ext))
+                # Rename the file so it has .prejbig2.ext extension
+                # Making it unique avoids problems with Windows if the
+                # same image is extracted multiple times
+                imgname.rename(imgname.with_suffix(".prejbig2" + ext))
+            except NotImplementedError as e:
+                if '/Decode' in str(e):
+                    log.debug(
+                        f"xref {xref}: skipping image with unsupported Decode table"
+                    )
+                    return None
+                raise
             except UnsupportedImageTypeError:
                 return None
             finally:
@@ -163,10 +187,18 @@ def extract_image_jbig2(
                     pim.obj.ColorSpace = colorspace
                 else:
                     del pim.obj.ColorSpace
-            return XrefExt(xref, ext)
+            return XrefExt(xref, ".prejbig2" + ext)
     return None
 
 
+def _should_optimize_jpeg(options, filtdp):
+    if options.optimize >= 2:
+        return True
+    # Ghostscript 10.6.0+ introduced some sort of JPEG encoding issue.
+    # To resolve this, re-optimize the JPEG anyway.
+    return options.optimize < 2 and ghostscript.version() >= Version('10.6.0')
+
+
 def extract_image_generic(
     *, pdf: Pdf, root: Path, image: Stream, xref: Xref, options
 ) -> XrefExt | None:
@@ -180,21 +212,13 @@ def extract_image_generic(
     if pim.bits_per_component == 1:
         return None
 
-    if filtdp[0] == Name.DCTDecode and options.optimize >= 2:
-        # This is a simple heuristic derived from some training data, that has
-        # about a 70% chance of guessing whether the JPEG is high quality,
-        # and possibly recompressible, or not. The number itself doesn't mean
-        # anything.
-        # bytes_per_pixel = int(raw_jpeg.Length) / (w * h)
-        # jpeg_quality_estimate = 117.0 * (bytes_per_pixel ** 0.213)
-        # if jpeg_quality_estimate < 65:
-        #     return None
+    if filtdp[0] == Name.DCTDecode and _should_optimize_jpeg(options, filtdp):
         try:
             imgname = root / f'{xref:08d}'
             with imgname.open('wb') as f:
                 ext = pim.extract_to(stream=f)
             imgname.rename(imgname.with_suffix(ext))
-        except UnsupportedImageTypeError:
+        except (UnsupportedImageTypeError, HifiPrintImageNotTranscodableError):
             return None
         return XrefExt(xref, ext)
     elif (
@@ -219,11 +243,9 @@ def extract_image_generic(
         not pim.indexed
         and pim.colorspace == Name.ICCBased
         and pim.bits_per_component == 1
-        and not options.jbig2_lossy
     ):
         # We can losslessly optimize 1-bit images to CCITT or JBIG2 without
-        # paying any attention to the ICC profile, provided we're not doing
-        # lossy JBIG2
+        # paying any attention to the ICC profile
         pim.as_pil_image().save(png_name(root, xref))
         return XrefExt(xref, '.png')
 
@@ -250,6 +272,9 @@ def _find_image_xrefs_container(
     for _imname, image in dict(xobjs).items():
         if image.objgen[1] != 0:
             continue  # Ignore images in an incremental PDF
+        xref = Xref(image.objgen[0])
+        if xref in include_xrefs or xref in exclude_xrefs:
+            continue  # Already processed
         if Name.Subtype in image and image.Subtype == Name.Form:
             # Recurse into Form XObjects
             log.debug(f"Recursing into Form XObject {_imname} in page {pageno}")
@@ -263,7 +288,6 @@ def _find_image_xrefs_container(
                 depth + 1,
             )
             continue
-        xref = Xref(image.objgen[0])
         if Name.SMask in image:
             # Ignore soft masks
             smask_xref = Xref(image.SMask.objgen[0])
@@ -344,116 +368,75 @@ def extract_images_generic(
     return jpegs, pngs
 
 
-def extract_images_jbig2(pdf: Pdf, root: Path, options) -> dict[int, list[XrefExt]]:
+def extract_images_jbig2(pdf: Pdf, root: Path, options) -> list[XrefExt]:
     """Extract any bitonal image that we think we can improve as JBIG2."""
-    jbig2_groups = defaultdict(list)
-    for pageno, xref_ext in extract_images(pdf, root, options, extract_image_jbig2):
-        group = pageno // options.jbig2_page_group_size
-        jbig2_groups[group].append(xref_ext)
+    jbig2_images = []
+    for _pageno, xref_ext in extract_images(pdf, root, options, extract_image_jbig2):
+        jbig2_images.append(xref_ext)
 
-    log.debug(f"Optimizable images: JBIG2 groups: {len(jbig2_groups)}")
-    return jbig2_groups
+    log.debug(f"Optimizable images: JBIG2: {len(jbig2_images)}")
+    return jbig2_images
 
 
 def _produce_jbig2_images(
-    jbig2_groups: dict[int, list[XrefExt]], root: Path, options, executor: Executor
+    jbig2_images: list[XrefExt], root: Path, options, executor: Executor
 ) -> None:
-    """Produce JBIG2 images from their groups."""
+    """Produce JBIG2 images using lossless single-image encoding."""
 
-    def jbig2_group_args(root: Path, groups: dict[int, list[XrefExt]]):
-        for group, xref_exts in groups.items():
-            prefix = f'group{group:08d}'
+    def jbig2_args():
+        for xref_ext in jbig2_images:
+            xref, ext = xref_ext
             yield (
-                fspath(root),  # =cwd
-                (img_name(root, xref, ext) for xref, ext in xref_exts),  # =infiles
-                prefix,  # =out_prefix
+                fspath(root),
+                img_name(root, xref, ext),
+                root / f'{xref:08d}.jbig2',
                 options.jbig2_threshold,
             )
 
-    def jbig2_single_args(root: Path, groups: dict[int, list[XrefExt]]):
-        for group, xref_exts in groups.items():
-            prefix = f'group{group:08d}'
-            # Second loop is to ensure multiple images per page are unpacked
-            for n, xref_ext in enumerate(xref_exts):
-                xref, ext = xref_ext
-                yield (
-                    fspath(root),
-                    img_name(root, xref, ext),
-                    root / f'{prefix}.{n:04d}',
-                    options.jbig2_threshold,
-                )
-
-    if options.jbig2_page_group_size > 1:
-        jbig2_args = jbig2_group_args
-        jbig2_convert = jbig2enc.convert_group
-    else:
-        jbig2_args = jbig2_single_args
-        jbig2_convert = jbig2enc.convert_single
-
     executor(
         use_threads=True,
         max_workers=options.jobs,
         progress_kwargs=dict(
-            total=len(jbig2_groups),
+            total=len(jbig2_images),
             desc="JBIG2",
-            unit='item',
+            unit='image',
             disable=not options.progress_bar,
         ),
-        task=jbig2_convert,
-        task_arguments=jbig2_args(root, jbig2_groups),
+        task=jbig2enc.convert_single,
+        task_arguments=jbig2_args(),
     )
 
 
 def convert_to_jbig2(
     pdf: Pdf,
-    jbig2_groups: dict[int, list[XrefExt]],
+    jbig2_images: list[XrefExt],
     root: Path,
     options,
     executor: Executor,
 ) -> None:
     """Convert images to JBIG2 and insert into PDF.
 
-    When the JBIG2 page group size is > 1 we do several JBIG2 images at once
-    and build a symbol dictionary that will span several pages. Each JBIG2
-    image must reference to its symbol dictionary. If too many pages shared the
-    same dictionary JBIG2 encoding becomes more expensive and less efficient.
-    The default value of 10 was determined through testing. Currently this
-    must be lossy encoding since jbig2enc does not support refinement coding.
-
-    When the JBIG2 symbolic coder is not used, each JBIG2 stands on its own
-    and needs no dictionary. Currently this must be lossless JBIG2.
+    Each JBIG2 image is encoded independently using lossless compression.
+    No symbol dictionary (JBIG2Globals) is used.
     """
-    jbig2_globals_dict: Dictionary | None
+    _produce_jbig2_images(jbig2_images, root, options, executor)
 
-    _produce_jbig2_images(jbig2_groups, root, options, executor)
-
-    for group, xref_exts in jbig2_groups.items():
-        prefix = f'group{group:08d}'
-        jbig2_symfile = root / (prefix + '.sym')
-        if jbig2_symfile.exists():
-            jbig2_globals_data = jbig2_symfile.read_bytes()
-            jbig2_globals = Stream(pdf, jbig2_globals_data)
-            jbig2_globals_dict = Dictionary(JBIG2Globals=jbig2_globals)
-        elif options.jbig2_page_group_size == 1:
-            jbig2_globals_dict = None
-        else:
-            raise FileNotFoundError(jbig2_symfile)
-
-        for n, xref_ext in enumerate(xref_exts):
-            xref, _ = xref_ext
-            jbig2_im_file = root / (prefix + f'.{n:04d}')
-            jbig2_im_data = jbig2_im_file.read_bytes()
-            im_obj = pdf.get_object(xref, 0)
-            im_obj.write(
-                jbig2_im_data, filter=Name.JBIG2Decode, decode_parms=jbig2_globals_dict
-            )
+    for xref_ext in jbig2_images:
+        xref, _ = xref_ext
+        jbig2_im_file = root / f'{xref:08d}.jbig2'
+        jbig2_im_data = jbig2_im_file.read_bytes()
+        im_obj = pdf.get_object(xref, 0)
+        im_obj.write(jbig2_im_data, filter=Name.JBIG2Decode, decode_parms=None)
 
 
 def _optimize_jpeg(
-    xref: Xref, in_jpg: Path, opt_jpg: Path, jpeg_quality: int
+    xref: Xref, in_jpg: Path, opt_jpg: Path, jpg_quality: int
 ) -> tuple[Xref, Path | None]:
     with Image.open(in_jpg) as im:
-        im.save(opt_jpg, optimize=True, quality=jpeg_quality)
+        save_kwargs: dict[str, Any] = {'optimize': True}
+        if isinstance(jpg_quality, int) and 0 < jpg_quality <= 100:
+            save_kwargs['quality'] = jpg_quality
+        im.save(opt_jpg, **save_kwargs)
 
     if opt_jpg.stat().st_size > in_jpg.stat().st_size:
         log.debug(f"xref {xref}, jpeg, made larger - skip")
@@ -471,7 +454,7 @@ def transcode_jpegs(
         for xref in jpegs:
             in_jpg = jpg_name(root, xref)
             opt_jpg = in_jpg.with_suffix('.opt.jpg')
-            yield xref, in_jpg, opt_jpg, options.jpeg_quality
+            yield xref, in_jpg, opt_jpg, options.jpg_quality
 
     def finish_jpeg(result: tuple[Xref, Path | None], pbar: ProgressBar):
         xref, opt_jpg = result
@@ -496,6 +479,16 @@ def transcode_jpegs(
     )
 
 
+def _already_flate_encoded(image: Stream) -> bool:
+    """Check if the image already has FlateDecode in its filter chain."""
+    filt = image.get(Name.Filter)
+    if filt is None:
+        return False
+    if isinstance(filt, Array):
+        return Name.FlateDecode in list(filt)
+    return filt == Name.FlateDecode
+
+
 def _find_deflatable_jpeg(
     *, pdf: Pdf, root: Path, image: Stream, xref: Xref, options
 ) -> XrefExt | None:
@@ -504,7 +497,23 @@ def _find_deflatable_jpeg(
         return None
     _pim, filtdp = result
 
-    if filtdp[0] == Name.DCTDecode and not filtdp[1] and options.optimize >= 1:
+    # Skip if already FlateDecode compressed - would double-compress
+    if _already_flate_encoded(image):
+        return None
+
+    if (
+        filtdp[0] == Name.DCTDecode
+        and not filtdp[1]
+        and (
+            (
+                # Don't flate very large images because it will slow down PDF viewers
+                1 <= options.optimize <= 2
+                and image.get(Name.Width, 0) < FLATE_JPEG_THRESHOLD
+                and image.get(Name.Height, 0) < FLATE_JPEG_THRESHOLD
+            )
+            or options.optimize == 3
+        )
+    ):
         return XrefExt(xref, '.memory')
 
     return None
@@ -672,12 +681,10 @@ def optimize(
         safe_symlink(input_file, output_file)
         return output_file
 
-    if options.jpeg_quality == 0:
-        options.jpeg_quality = DEFAULT_JPEG_QUALITY if options.optimize < 3 else 40
+    if options.jpg_quality == 0:
+        options.jpg_quality = DEFAULT_JPEG_QUALITY if options.optimize < 3 else 40
     if options.png_quality == 0:
         options.png_quality = DEFAULT_PNG_QUALITY if options.optimize < 3 else 30
-    if options.jbig2_page_group_size == 0:
-        options.jbig2_page_group_size = 10 if options.jbig2_lossy else 1
 
     with Pdf.open(input_file) as pdf:
         root = output_file.parent / 'images'
@@ -691,8 +698,8 @@ def optimize(
         #    transcode_pngs(pdf, jpegs, jpg_name, root, options)
         transcode_pngs(pdf, pngs, png_name, root, options, executor)
 
-        jbig2_groups = extract_images_jbig2(pdf, root, options)
-        convert_to_jbig2(pdf, jbig2_groups, root, options, executor)
+        jbig2_images = extract_images_jbig2(pdf, root, options)
+        convert_to_jbig2(pdf, jbig2_images, root, options, executor)
 
         target_file = output_file.with_suffix('.opt.pdf')
         pdf.remove_unreferenced_resources()
@@ -727,35 +734,25 @@ def main(infile, outfile, level, jobs=1):
     from shutil import copy  # pylint: disable=import-outside-toplevel
     from tempfile import TemporaryDirectory  # pylint: disable=import-outside-toplevel
 
-    class OptimizeOptions:
-        """Emulate ocrmypdf's options."""
-
-        def __init__(
-            self, input_file, jobs, optimize_, jpeg_quality, png_quality, jb2lossy
-        ):
-            self.input_file = input_file
-            self.jobs = jobs
-            self.optimize = optimize_
-            self.jpeg_quality = jpeg_quality
-            self.png_quality = png_quality
-            self.jbig2_page_group_size = 0
-            self.jbig2_lossy = jb2lossy
-            self.jbig2_threshold = 0.85
-            self.quiet = True
-            self.progress_bar = False
+    from ocrmypdf._options import OcrOptions  # pylint: disable=import-outside-toplevel
 
     infile = Path(infile)
-    options = OptimizeOptions(
+
+    # Create OcrOptions with optimization-specific settings
+    options = OcrOptions(
         input_file=infile,
+        output_file=outfile,  # Required field
         jobs=jobs,
-        optimize_=int(level),
-        jpeg_quality=0,  # Use default
+        optimize=int(level),
+        jpg_quality=0,  # Use default
         png_quality=0,
-        jb2lossy=False,
+        jbig2_threshold=0.85,
+        quiet=True,
+        progress_bar=False,
     )
 
     with TemporaryDirectory() as tmpdir:
-        context = PdfContext(options, tmpdir, infile, None, None)
+        context = PdfContext(options, Path(tmpdir), infile, None, None)
         tmpout = Path(tmpdir) / 'out.pdf'
         optimize(
             infile,
